//
// Copyright 2020 Ettus Research, a National Instruments Brand
//
// SPDX-License-Identifier: GPL-3.0-or-later
//

#pragma once

namespace uhd { namespace math {

enum class interp_mode { NEAREST_NEIGHBOR, LINEAR };

}} // namespace uhd::math
